#include <QCoreApplication>
#include <QDebug>

#include <oapi_client/ClientToServerApi.h>

using Id = qint64;

void PrintSource(const oapi::SourceInfo& source)
{
    qDebug() << "id" << source.id << "name"
             << source.name
             // converting enums to string
             << "type" << oapi::toJson(source.type) << "state" << oapi::toJson(source.state);
}

int main(int argc, char** argv)
{
    QCoreApplication app(argc, argv);
    qDebug() << "Subscribe to sources list using long polling";

    // Create api instance, set radar_server address & timeout
    auto server_api = QSharedPointer<oapi::ClientToServerApi>::create();
    server_api->setHttpConnectionSettings(utils::NetworkAddress{QHostAddress("127.0.0.1"), 3000});

    auto subscription =
        QSharedPointer<oapi::GetSourcesInfoSubscription>(new oapi::GetSourcesInfoSubscription());

    std::map<Id, oapi::SourceInfo> local_sources;

    // Subscribe response
    QObject::connect(
        subscription.get(),
        &oapi::GetSourcesInfoSubscription::onResponse,
        &app,
        [s = subscription.get(), &local_sources]()
        {
            if (s->error())
            {
                qDebug() << s->error()->description;
                return;
            }

            if (s->lastModifiedTime())
            {
                qDebug() << "Last sources modification time:"
                         << s->lastModifiedTime()->toString(Qt::ISODateWithMs);
            }

            auto& response = *s->response();

            if (!response.all_sources)
            {
                qDebug() << "Received full list of sources on first request: ";

                local_sources.clear();
                for (auto&& source : response.sources)
                {
                    PrintSource(source);
                    local_sources.emplace(source.id, std::move(source));
                }

                qDebug() << "";
            }
            else
            {
                // remove local sources that do not exist on the server anymore
                const auto& all_sources = *response.all_sources;
                auto source_iter = local_sources.begin();
                while (source_iter != local_sources.end())
                {
                    if (std::find(all_sources.begin(), all_sources.end(), source_iter->first) ==
                        all_sources.end())
                    {
                        qDebug() << "Source removed:";
                        PrintSource(source_iter->second);
                        source_iter = local_sources.erase(source_iter);
                    }
                    else
                    {
                        ++source_iter;
                    }
                }

                // create/update local sources
                for (auto&& source : response.sources)
                {
                    auto item = local_sources.insert_or_assign(source.id, std::move(source));
                    qDebug() << QString("Source %1:").arg(item.second ? "created" : "updated");
                    PrintSource(item.first->second);
                }

                qDebug() << "";

                const oapi::SourcesInfo& sources = *s->response();
                qDebug() << oapi::toJsonString(sources);
            }
        });

    subscription->setRequest(
        [s = subscription.get(), api = server_api.get()](QObject* owner, auto on_response) {
            api->getSourcesInfo(
                s->lastModifiedTime(), owner, on_response, http_client::kLongPollTimeout);
        });

    return app.exec();
}
