#include <QCoreApplication>

#include <oapi_client/ClientToServerApi.h>
#include <oapi_client/ConfigurationApi.h>
#include <oapi_models/SourceLocationParameters.h>

using namespace utils;

void SetRadarLocation(QObject* owner,
                      oapi::ConfigurationApi& source_api,
                      double latitude,
                      double longitude);

int main(int argc, char* argv[])
{
    QCoreApplication app(argc, argv);

    oapi::ClientToServerApi api;

    auto server_address = utils::NetworkAddress{QHostAddress("127.0.0.1"), 3000};

    api.setHttpConnectionSettings(server_address);

    oapi::ConfigurationApi source_api;

    api.getSourcesInfo(std::nullopt,
                       &app,
                       [&app, &source_api](RequestResult<oapi::SourcesInfo> result,
                                           http_client::NetworkReplyData /*network_reply_data*/)
                       {
                           if (result.error)
                           {
                               qDebug()
                                   << "Could not read sources: " << result.error->network_error;
                               return;
                           }

                           const auto& sources = result.result->sources;

                           for (const auto& source : sources)
                           {
                               if (source.type == oapi::SourceType::E_ENOT_RADAR)
                               {
                                   qDebug() << "Setting location for " << oapi::toJson(source.type)
                                            << "source" << source.id;

                                   source_api.setHttpConnectionSettings(utils::NetworkAddress(
                                       QHostAddress(source.address), source.port));

                                   SetRadarLocation(&app, source_api, 54.3354, 34.56775);
                                   return;
                               }
                           }
                           qDebug() << "No enot-radar sources";
                           app.quit();
                       });

    return app.exec();
}

void SetRadarLocation(QObject* owner,
                      oapi::ConfigurationApi& source_api,
                      double latitude,
                      double longitude)
{
    oapi::ConfigurationGroupValues location_settings;
    oapi::Value val_latitude;
    val_latitude.data = QString::number(latitude);
    val_latitude.type = oapi::ConfigurationItemType::E_NUMBER;

    oapi::Value val_longitude;
    val_longitude.data = QString::number(longitude);
    val_longitude.type = oapi::ConfigurationItemType::E_NUMBER;

    location_settings.values.emplace(oapi::SourceLocationParametersStrings::E_LATITUDE,
                                     std::move(val_latitude));
    location_settings.values.emplace(oapi::SourceLocationParametersStrings::E_LONGITUDE,
                                     std::move(val_longitude));

    oapi::ConfigurationValues groups;
    groups.groups.emplace(oapi::SourceLocationParametersStrings::E_THIS_ENUM,  // group name
                          std::move(location_settings));

    source_api.setConfiguration(
        groups,
        owner,
        [](ResultOptional<QString, RequestError> response, http_client::NetworkReplyData)
        {
            if (response.error)
            {
                qDebug().noquote() << "Configuraion was not applied:";
                qDebug().noquote() << response.error->description;
            }
            else
            {
                qDebug().noquote() << "Configuration applied:";
                qDebug().noquote() << *response.result;
            }
        });
}
