#include <QCoreApplication>

#include <oapi_client/ClientToServerApi.h>
#include <oapi_models/CommonFrameType.h>
#include <oapi_models/EnotIntensityMapFormat.h>
#include <oapi_models/EnotIntensityMapFrame.h>
#include <oapi_models/TrajectoryChangesFrame.h>
#include <oapi_models/TrajectoryFrameType.h>
#include <oapi_models/TrajectoryInterpolationFrame.h>
#include <oapi_models/TrajectoryMergerFrame.h>
#include <streaming_api/stream.h>
#include <utils/type_helpers.h>

using namespace utils;

void InitStream(QObject* owner,
                const utils::NetworkAddress& server_address,
                const std::optional<qint64> source_id);

void ProcessMessage(const QByteArray& message);

int main(int argc, char* argv[])
{
    QCoreApplication app(argc, argv);

    oapi::ClientToServerApi api;

    auto server_address = utils::NetworkAddress{QHostAddress("127.0.0.1"), 3000};

    api.setHttpConnectionSettings(server_address);

    api.getSourcesInfo(
        std::nullopt,
        &app,
        [&app, server_address](RequestResult<oapi::SourcesInfo> result,
                               http_client::NetworkReplyData /*network_reply_data*/)
        {
            if (result.error)
            {
                qDebug() << "Could not read sources: " << result.error->network_error;
                return;
            }

            const auto& sources = result.result->sources;

            for (const auto& source : sources)
            {
                if (source.type == oapi::SourceType::E_ENOT_RADAR)
                {
                    qDebug() << "Reading streams for" << oapi::toJson(source.type) << "source"
                             << source.id;

                    // todo: refactoring
                    InitStream(&app, server_address, source.id);  // reading source stream
                    InitStream(
                        &app,
                        server_address,
                        std::nullopt);  // reading server stream (TrajectoryFrameType::E_MERGER)
                                        // when source_id is not set
                    return;
                }
            }
            qDebug() << "No enot-radar sources";
            app.quit();
        });

    return app.exec();
}

void InitStream(QObject* owner,
                const utils::NetworkAddress& server_address,
                const std::optional<qint64> source_id)
{
    auto stream = new streaming_api::Stream(owner);

    // streaming port is http port + 1
    auto streams_address = server_address;
    streams_address.port += 1;

    stream->open(streams_address);

    QObject::connect(
        stream,
        &streaming_api::Stream::stateChanged,
        owner,
        [](QAbstractSocket::SocketState state)
        { qDebug() << "Stream connection state (QAbstractSocket::SocketState)" << state; });

    QObject::connect(
        stream, &streaming_api::Stream::connected, owner, [] { qDebug() << "Stream connected"; });

    QObject::connect(stream, &streaming_api::Stream::onMessage, owner, &ProcessMessage);

    OAISet<oapi::ReadStreamConfig> streams = {{oapi::StreamType::E_TRAJECTORIES, std::nullopt}};
    if (source_id)
    {
        streams.insert({oapi::StreamType::E_ENOT_INTENSITY_MAP, std::nullopt});
    }

    stream->initRead({source_id, std::nullopt, std::nullopt});
    stream->readStreams(streams);
}

void ProcessMessage(const QByteArray& message)
{
    const streaming_api::Header* header = streaming_api::GetHeader(message);

    switch (header->stream_type)
    {
    case oapi::StreamType::E_TRAJECTORIES:
    {
        if (header->frame_type == utils::to_underlying(oapi::TrajectoryFrameType::E_CHANGES))
        {
            oapi::TrajectoryChangesFrame frame;
            auto error = streaming_api::Deserialize(frame, message);
            if (error)
            {
                qDebug() << "TrajectoryChangesFrame deserialization error:" << *error;
                return;
            }

            for (auto& p : frame.new_points)
            {
                if (!p.enot_data)
                {
                    return;
                }

                qDebug() << "Trajectory" << p.id.track_id
                         << "updated, position:" << toJson(p.enot_data->position.position);
            }
        }
        else if (header->frame_type ==
                 utils::to_underlying(oapi::TrajectoryFrameType::E_INTERPOLATION))
        {
            oapi::TrajectoryInterpolationFrame frame;
            auto error = streaming_api::Deserialize(frame, message);
            if (error)
            {
                qDebug() << "TrajectoryInterpolationFrame deserialization error:" << *error;
                return;
            }

            for (auto& p : frame.interpolation)
            {
                if (!p.rewrite)
                {
                    qDebug() << "Trajectory" << p.id.track_id
                             << "interpolated position updated, last point:"
                             << toJson(p.positions.back());
                }
            }
        }
        else if (header->frame_type == utils::to_underlying(oapi::TrajectoryFrameType::E_MERGER))
        {
            oapi::TrajectoryMergerFrame frame;
            auto error = streaming_api::Deserialize(frame, message);
            if (error)
            {
                qDebug() << "TrajectoryMergerFrame deserialization error:" << *error;
                return;
            }

            for (auto& g : frame.new_groups)
            {
                qDebug() << QString("Group of trajectories #%1 (%2) with main track %3")
                                .arg(QString::number(g.id),
                                     toJsonString(g.tracks),
                                     toJsonString(g.main_track));
            }
        }
        break;
    }

    case oapi::StreamType::E_ENOT_INTENSITY_MAP:
    {
        if (header->frame_type == utils::to_underlying(oapi::CommonFrameType::E_FORMAT))
        {
            oapi::EnotIntensityMapFormat frame;
            auto error = streaming_api::Deserialize(frame, message);
            if (error)
            {
                qDebug() << "EnotIntensityMapFormat deserialization error:" << *error;
                return;
            }
            // todo: save format
            qDebug() << oapi::toJson(frame);
        }
        else if (header->frame_type == utils::to_underlying(oapi::CommonFrameType::E_DATA))
        {
            oapi::EnotIntensityMapFrame frame;
            auto error = streaming_api::Deserialize(frame, message);
            if (error)
            {
                qDebug() << "EnotIntensityMapFrame deserialization error:" << *error;
                return;
            }

            // todo: use saved format
            qDebug() << "intensity map azimuth:" << frame.azimuth;
        }

        break;
    }

    default:
        break;
    };
};
