#include <QCoreApplication>

#include <oapi_client/ClientToServerApi.h>
#include <oapi_models/CommonFrameType.h>
#include <oapi_models/ReadStreamPlaybackState.h>
#include <oapi_models/TrajectoryChangesFrame.h>
#include <oapi_models/TrajectoryFrameType.h>
#include <streaming_api/stream.h>
#include <utils/type_helpers.h>

using namespace utils;

void InitStream(QCoreApplication* app,
                const utils::NetworkAddress& server_address,
                const std::optional<qint64> source_id);

void ProcessMessage(const QByteArray& message, streaming_api::Stream* stream);

int main(int argc, char* argv[])
{
    QCoreApplication app(argc, argv);

    oapi::ClientToServerApi api;

    auto server_address = utils::NetworkAddress{QHostAddress("127.0.0.1"), 3000};

    api.setHttpConnectionSettings(server_address);

    api.getSourcesInfo(std::nullopt,
                       &app,
                       [&app, server_address](RequestResult<oapi::SourcesInfo> result,
                                              http_client::NetworkReplyData /*network_reply_data*/)
                       {
                           if (result.error)
                           {
                               qDebug()
                                   << "Could not read sources: " << result.error->network_error;
                               return;
                           }

                           const auto& sources = result.result->sources;

                           for (const auto& source : sources)
                           {
                               if (source.type == oapi::SourceType::E_ENOT_RADAR)
                               {
                                   qDebug() << "Reading streams for" << oapi::toJson(source.type)
                                            << "source" << source.id;

                                   InitStream(&app, server_address, source.id);
                                   return;
                               }
                           }

                           qDebug() << "No enot-radar sources";
                           app.quit();
                       });

    return app.exec();
}

void InitStream(QCoreApplication* app,
                const utils::NetworkAddress& server_address,
                const std::optional<qint64> source_id)
{
    auto stream = new streaming_api::Stream(app);

    // streaming port is http port + 1
    auto streams_address = server_address;
    streams_address.port += 1;

    stream->open(streams_address);

    QObject::connect(
        stream,
        &streaming_api::Stream::stateChanged,
        app,
        [](QAbstractSocket::SocketState state)
        { qDebug() << "Stream connection state (QAbstractSocket::SocketState)" << state; });

    QObject::connect(
        stream, &streaming_api::Stream::connected, app, [] { qDebug() << "Stream connected"; });

    QObject::connect(stream,
                     &streaming_api::Stream::onMessage,
                     app,
                     [stream](const QByteArray& message) { ProcessMessage(message, stream); });

    const int kExportDepthSecs = 60;

    stream->initRead({source_id, std::nullopt, std::nullopt});
    stream->readStreams({{oapi::StreamType::E_TRAJECTORIES, kExportDepthSecs}});
    stream->playback()->setSpeed(0.0);  // do not continue playing after export

    // export one minute from two minutes ago (from -120s to -60s].
    // stream->playback()->seek(QDateTime::currentDateTimeUtc().addSecs(-60));
    stream->playback()->seekToRealtime();
    // or seekToRealtime() for export last 60 seconds of archive

    QObject::connect(stream, &streaming_api::Stream::disconnected, app, &QCoreApplication::quit);
    QObject::connect(stream, &QObject::destroyed, app, &QCoreApplication::quit);
}

void ProcessMessage(const QByteArray& message, streaming_api::Stream* stream)
{
    static size_t frames = 0;

    const streaming_api::Header* header = streaming_api::GetHeader(message);

    switch (header->stream_type)
    {
    case oapi::StreamType::E_STREAM_CONFIG:
    {
        oapi::ReadStreamPlaybackState state;
        auto error = streaming_api::Deserialize(state, message);
        if (error)
        {
            qDebug() << "ReadStreamPlaybackState deserialization error:" << *error;
            return;
        }

        if (state.end_of_chunk == oapi::StreamType::E_TRAJECTORIES)
        {
            // Archive chunk finished. You may need to save last output file.
            qDebug() << "Archive chunk finished, chunk frames:" << frames;
            frames = 0;
        }

        if (state.end_of_export == oapi::StreamType::E_TRAJECTORIES)
        {
            // Export finished. You may need to save last output file.

            // If stream is paused (playback speed is 0), there will be no more
            // frames. Otherwise playback will continue starting from seek time.
            qDebug() << "Export finished, last chunk frames:" << frames;
            stream->deleteLater();
        }
        break;
    }
    case oapi::StreamType::E_TRAJECTORIES:
    {
        if (header->frame_type == utils::to_underlying(oapi::TrajectoryFrameType::E_CHANGES))
        {
            oapi::TrajectoryChangesFrame frame;
            auto error = streaming_api::Deserialize(frame, message);
            if (error)
            {
                qDebug() << "TrajectoryChangesFrame deserialization error:" << *error;
                return;
            }

            qDebug() << "frame :" << frame.time.toString(Qt::ISODateWithMs);
            frames++;

            // Create last output file if not exist. Append data.
        }
        break;
    default:
        break;
    };
    };
}
