#include <QCoreApplication>
#include <QDebug>
#include <QDir>
#include <QElapsedTimer>
#include <memory>

#include <oapi_models/CommonFrameType.h>
#include <oapi_models/EnotIntensityMapFrame.h>
#include <oapi_models/EnotSpectrumFrame.h>
#include <oapi_models/TrajectoryPoint.h>
#include <streaming_api/frame.h>
#include <utils/type_helpers.h>

oapi::EnotSpectrumFrame CreateSpectrum()
{
    oapi::EnotSpectrumFrame s;
    s.format.azimuth = 12;
    s.format.azimuth_size = 34;
    s.format.range_size = 4096 * 10;
    s.format.range_step = 78;
    s.format.speed_size = 64;
    s.format.speed_step = 91;
    s.format.start_range = 92;
    s.max_incremental_azimuth = 98;
    s.min_incremental_azimuth = 76;
    s.mask.resize(s.format.range_size * s.format.speed_size);
    for (auto& v : s.mask)
    {
        v = rand() % 255;
    }

    s.power.resize(s.format.range_size * s.format.speed_size);
    for (auto& v : s.power)
    {
        v = rand() / 1000.f;
    }

    return s;
}

int main(int argc, char** argv)
{
    QCoreApplication app(argc, argv);

    using namespace oapi;

    std::vector<QByteArray> frames;
    std::vector<streaming_api::Header> headers;

    {
        auto header = streaming_api::Header(StreamType::E_ENOT_SPECTRUM_DATA,
                                            utils::to_underlying(CommonFrameType::E_DATA),
                                            QDateTime::currentDateTimeUtc().toMSecsSinceEpoch(),
                                            false,
                                            SerializationFormat::E_JSON);

        headers.push_back(header);
        header.gzip_compressed = true;
        headers.push_back(header);
        header.gzip_compressed = false;
        header.serialization_format = SerializationFormat::E_CBOR;
        headers.push_back(header);
        header.gzip_compressed = true;
        headers.push_back(header);
    }

    qDebug() << "Serialization:";
    auto s = CreateSpectrum();

    QElapsedTimer timer;
    for (auto& h : headers)
    {
        timer.restart();
        frames.emplace_back(streaming_api::Serialize(s, h));
        qDebug() << "format" << toJson(h.serialization_format)
                 << (h.gzip_compressed ? "with gzip compression" : "") << "size"
                 << frames.back().size() << "time" << timer.elapsed();
    }

    qDebug() << "---";

    qDebug() << "Deserialization:";
    std::vector<EnotSpectrumFrame> decoded_structs;
    std::vector<std::optional<Error>> errors;
    decoded_structs.resize(frames.size());
    errors.resize(frames.size());

    for (int i = 0; i < frames.size(); ++i)
    {
        auto h = streaming_api::GetHeader(frames[i]);

        if (!h)
        {
            qDebug() << "could not parse header";
            continue;
        }

        timer.restart();
        errors[i] = streaming_api::Deserialize(decoded_structs[i], frames[i]);

        qDebug() << "format" << toJson(h->serialization_format)
                 << (h->gzip_compressed ? "with gzip compression" : "") << "size"
                 << frames[i].size() << "time" << timer.elapsed();

        if (errors[i])
        {
            qDebug() << *errors[i];
        }
    }

    return app.exec();
}
