#ifndef BASE_API_H
#define BASE_API_H

#include <QNetworkReply>
#include <QObject>

#include <utils/result.h>

#include "http_requests.h"
#include "requests_manager.h"

namespace http_client {

class BaseApi : public QObject
{
    Q_OBJECT
public:
    explicit BaseApi(QObject* parent = nullptr, const QString& user_agent = "");

    void setHttpConnectionSettings(const utils::NetworkAddress& http_address,
                                   uint32_t timeout = kMinTimeout);
    const utils::NetworkAddress& getServerHttpAddress() const;

protected:
    utils::NetworkAddress m_http_address;
    const QString m_user_agent;

    RequestsManager m_network_manager;
    uint32_t m_timeout = kMinTimeout;
};

class BaseRequestSubscription : public QObject
{
    Q_OBJECT
public:
    BaseRequestSubscription(QObject* parent = nullptr) : QObject(parent) {}

signals:
    void onResponse();
};

template <typename T>
using ReplyCallback =
    std::function<void(ResultOptional<T, RequestError>, http_client::NetworkReplyData)>;

template <typename T>
using RequestCallback = std::function<void(QObject* owner, ReplyCallback<T> on_response)>;

template <typename T>
class RequestSubscription : public BaseRequestSubscription
{
private:
    static const int kMinimalRequestPeriodMs = 25;

public:
    RequestSubscription(QObject* parent = nullptr)
            : BaseRequestSubscription(parent)
            , m_request_owner(QSharedPointer<QObject>(new QObject()))
    {
    }

    void setRequest(RequestCallback<T> request_callback, int period_ms = 1000)
    {
        m_request_period_ms = period_ms;

        m_request_timer.disconnect();
        cancelLastRequest();
        resetLastModified();

        m_last_request_end = QDateTime::currentDateTimeUtc().addMSecs(-m_request_period_ms);
        m_last_response = RequestError("not_initialized", QNetworkReply::NetworkError::NoError, 0);

        QObject::connect(
            &m_request_timer,
            &QTimer::timeout,
            this,
            [this, request_callback]
            {
                if (!m_last_request_end.isValid() ||
                    m_last_request_end.msecsTo(QDateTime::currentDateTimeUtc()) <
                        m_request_period_ms)
                {
                    return;
                }

                m_last_request_start = QDateTime::currentDateTimeUtc();
                m_last_request_end = QDateTime();

                request_callback(
                    m_request_owner.get(),
                    [this](ResultOptional<T, RequestError> response,
                           http_client::NetworkReplyData data)
                    {
                        m_last_request_end = QDateTime::currentDateTimeUtc();

                        if (data.error == QNetworkReply::OperationCanceledError)
                        {
                            return;  // long polling timeout, no error, no data was updated
                        }

                        m_last_response = std::move(response);
                        if (data.error == QNetworkReply::NetworkError::NoError)
                        {
                            m_last_modified_header = data.last_modified;
                        }

                        emit onResponse();
                    });
            });
        m_request_timer.start(kMinimalRequestPeriodMs);
    }

    void stop()
    {
        m_request_timer.stop();
    }

    void resume()
    {
        m_request_timer.start();
    }

    const std::optional<RequestError>& error() const
    {
        return m_last_response.error;
    }

    const std::optional<T>& response() const
    {
        return m_last_response.result;
    }

    const ResultOptional<T, RequestError>& responseData() const
    {
        return m_last_response;
    }

    const std::optional<QDateTime>& lastModifiedTime() const
    {
        return m_last_modified_header;
    }

    void cancelLastRequest()
    {
        m_request_owner = QSharedPointer<QObject>(new QObject());
    }

private:
    void resetLastModified()
    {
        m_last_modified_header.reset();
    }

    ResultOptional<T, RequestError> m_last_response =
        RequestError("not_initialized", QNetworkReply::NetworkError::NoError, 0);

    QTimer m_request_timer;
    QDateTime m_last_request_start;
    QDateTime m_last_request_end;
    int m_request_period_ms = 0;

    std::optional<QDateTime> m_last_modified_header;

    QSharedPointer<QObject> m_request_owner;
};

}  // namespace http_client

#endif
