#include "http_requests.h"

#include <QNetworkReply>
#include <QPointer>

#include <utils/converters.h>
#include <utils/gzip_compressor.h>
#include <utils/logger.h>

namespace http_client {

const int kMinCompressionBytes = 256;

QNetworkRequest CreateRequest(const utils::NetworkAddress& address,
                              const QString& request,
                              const std::map<QString, QVariant>& parameters,
                              const QString& user_agent,
                              const std::optional<http_server::ContentType>& content_type)
{
    return CreateRequest(http_server::kHTTPProtocol + address.toString() + request,
                         parameters,
                         user_agent,
                         content_type);
}

QNetworkRequest CreateRequest(const QString& request,
                              const std::map<QString, QVariant>& parameters,
                              const QString& user_agent,
                              const std::optional<http_server::ContentType>& content_type)
{
    QString request_str = request;
    if (!parameters.empty())
    {
        request_str += "?";
        for (auto p = parameters.begin(); p != parameters.end(); ++p)
        {
            QString val;

            if (p->second.type() == QVariant::Bool)
            {
                using namespace utils;
                val = utils::BoolToStr(p->second.toBool());
            }
            else
            {
                val = p->second.toString();
            }

            if (p != parameters.begin())
            {
                request_str += "&";
            }

            request_str += p->first + "=" + val;
        }
    }

    QNetworkRequest qrequest(request_str);
    qrequest.setRawHeader("User-Agent", user_agent.toUtf8());
    if (content_type)
    {
        qrequest.setHeader(QNetworkRequest::ContentTypeHeader,
                           http_server::kContentTypes.at(*content_type).toUtf8());
    }
    return qrequest;
}

// connect to owner's callback, cancel by timeout and automatically destroy reply object
void configureReply(QNetworkReply* reply,
                    QPointer<QObject> receiver,
                    const std::function<void(NetworkReplyData)>& on_reply)
{
    if (!reply || !on_reply || !receiver)
    {
        return;
    }

    QObject::connect(receiver, &QObject::destroyed, reply, &QNetworkReply::abort);

    // QObject::connect(reply, &QNetworkReply::errorOccurred, [] (QNetworkReply::NetworkError e)
    //{
    //     qDebug() << (int) e;
    // });

    QObject::connect(reply,
                     &QNetworkReply::finished,
                     receiver,
                     [on_reply, reply, receiver]
                     {
                         reply->deleteLater();

                         if (!receiver)
                         {
                             return;
                         }

                         auto last_modified = QDateTime::fromString(
                             reply->header(QNetworkRequest::ETagHeader).toString().remove('"'),
                             Qt::ISODateWithMs);

                         NetworkReplyData reply_data;
                         if (reply->bytesAvailable())
                         {
                             reply_data.data = reply->readAll();
                         }

                         reply_data.error = reply->error();
                         if (reply_data.error != QNetworkReply::NetworkError::NoError)
                         {
                             reply_data.error_string = reply->errorString();
                         }
                         reply_data.http_status_code =
                             reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();
                         reply_data.content_type = http_server::ContentTypeFromString(
                             reply->header(QNetworkRequest::ContentTypeHeader).toString());

                         if (last_modified.isValid())
                         {
                             reply_data.last_modified = last_modified;
                         }

                         on_reply(reply_data);
                     });
}

void AsyncRequest(RequestsManager& manager,
                  const http_server::RequestType type,
                  QNetworkRequest request,
                  QByteArray post_data,
                  const bool post_data_compress,
                  QObject* receiver,
                  const std::function<void(NetworkReplyData)>& on_reply,
                  int timeout)
{
    if (timeout < kMinTimeout)
    {
        // otherwise QNetworkReply cannot distinguish ConnectionRefusedError from
        // OperationCanceledError
        Log(QtWarningMsg,
            FileLogContext,
            QString("Timeouts less than %1ms are not supported").arg(QString::number(kMinTimeout)));
        timeout = kMinTimeout;
    }
    request.setTransferTimeout(timeout);

    if (type == http_server::RequestType::E_POST)
    {
        std::optional<QByteArray> compressed_post_data;

        if (post_data_compress && post_data.size() > kMinCompressionBytes)
        {
            compressed_post_data = QByteArray();
            if (!gzip::Compress(reinterpret_cast<const uint8_t*>(post_data.constData()),
                                post_data.size(),
                                *compressed_post_data))
            {
                Log(QtFatalMsg, FileLogContext, "AsyncPost() compression failed");
                return;
            }
            request.setRawHeader("Content-Encoding", "gzip");
        }

        configureReply(
            manager.post(request, compressed_post_data ? *compressed_post_data : post_data),
            receiver,
            on_reply);
    }
    else if (type == http_server::RequestType::E_GET)
    {
        configureReply(manager.get(request), receiver, on_reply);
    }
    else if (type == http_server::RequestType::E_DELETE)
    {
        configureReply(manager.deleteResource(request), receiver, on_reply);
    }
}

}  // namespace http_client
