#include "requests_manager.h"

#include <QNetworkReply>

#include <utils/utils.h>

namespace http_client {

RequestsManager::RequestsManager(QObject* parent) : QObject(parent)
{
    auto timer = new QTimer(this);
    QObject::connect(timer, &QTimer::timeout, this, &RequestsManager::removeUnused);
    timer->start(kTimeoutMs);
}

QNetworkReply* RequestsManager::get(const QNetworkRequest& request)
{
    return sendRequest(&QNetworkAccessManager::get, request);
}

QNetworkReply* RequestsManager::post(const QNetworkRequest& request, const QByteArray& data)
{
    return sendRequestWithBody(&QNetworkAccessManager::post, request, data);
}

QNetworkReply* RequestsManager::deleteResource(const QNetworkRequest& request)
{
    return sendRequest(&QNetworkAccessManager::deleteResource, request);
}

void RequestsManager::removeUnused()
{
    utils::erase_if(m_managers,
                    [this](auto& manager)
                    {
                        return manager.second.connection_count == 0 &&
                               manager.second.last_used.msecsTo(QDateTime::currentDateTime()) >
                                   kTimeoutMs;
                    });
}

QNetworkReply* RequestsManager::sendRequest(RequestMethod method, const QNetworkRequest& request)
{
    QNetworkAccessManager* manager = getManager();
    QNetworkReply* reply = (manager->*method)(request);
    handleReply(reply, manager);
    return reply;
}

QNetworkReply* RequestsManager::sendRequestWithBody(RequestMethodWithBody method,
                                                    const QNetworkRequest& request,
                                                    const QByteArray& data)
{
    QNetworkAccessManager* manager = getManager();
    QNetworkReply* reply = (manager->*method)(request, data);
    handleReply(reply, manager);
    return reply;
}

QNetworkAccessManager* RequestsManager::getManager()
{
    for (auto& manager : m_managers)
    {
        if (manager.second.connection_count < kMaxConnections)
        {
            manager.second.connection_count++;
            manager.second.last_used = QDateTime::currentDateTime();
            return manager.first;
        }
    }

    auto m = new QNetworkAccessManager(this);
    m_managers[m] = ManagerInfo{m, 1, QDateTime::currentDateTime()};
    return m;
}

void RequestsManager::handleReply(QNetworkReply* reply, QNetworkAccessManager* manager)
{
    connect(reply,
            &QNetworkReply::finished,
            this,
            [this, manager]()
            {
                auto iter = m_managers.find(manager);
                if (iter != m_managers.end())
                {
                    m_managers[manager].connection_count -= 1;
                }
            });
}

}  // namespace http_client
