#ifndef STREAMING_API_FRAME_H
#define STREAMING_API_FRAME_H

#include <QBuffer>
#include <QByteArray>

#include <oapi_models/SerializationFormat.h>
#include <oapi_models/StreamType.h>
#include <utils/gzip_compressor.h>
#include <utils/logger.h>

namespace streaming_api {

#ifdef OAPI_CBOR_SUPPORT
const auto kDefaultSerializationFormat = oapi::SerializationFormat::E_CBOR;
#else
const auto kDefaultSerializationFormat = oapi::SerializationFormat::E_JSON;
#endif

#pragma pack(push, 1)
struct Header
{
    oapi::StreamType stream_type;
    uint8_t frame_type;
    oapi::SerializationFormat serialization_format;
    bool gzip_compressed;
    int64_t time;  // utc server time when frame is sent

    Header(const oapi::StreamType _stream_type,
           const uint8_t _frame_type,
           int64_t _time = QDateTime::currentDateTimeUtc().toMSecsSinceEpoch(),
           bool _gzip_compression = true,
           const oapi::SerializationFormat _serialization_format = kDefaultSerializationFormat)
            : stream_type(_stream_type)
            , frame_type(_frame_type)
            , serialization_format(_serialization_format)
            , gzip_compressed(_gzip_compression)
            , time(_time)
    {
    }
};
#pragma pack(pop)

const Header* GetHeader(const QByteArray& frame);
Header* GetHeader(QByteArray& frame);

template <typename T>
QByteArray Serialize(T&& object, const Header& header)
{
    QByteArray a;
    QBuffer b(&a);
    b.open(QIODevice::Append);
    b.write(reinterpret_cast<const char*>(&header), sizeof(header));

    switch (header.serialization_format)
    {
    case oapi::SerializationFormat::E_JSON:
    {
        auto json_buffer = oapi::toJson(object);
        if (header.gzip_compressed)
        {
            QByteArray compressed;
            gzip::Compress(
                reinterpret_cast<uint8_t*>(json_buffer.data()), json_buffer.size(), compressed);
            b.write(compressed);
        }
        else
        {
            b.write(json_buffer);
        }
        break;
    }
    case oapi::SerializationFormat::E_CBOR:
    {
#ifdef OAPI_CBOR_SUPPORT
        if (header.gzip_compressed)
        {
            QByteArray serialized;
            QCborStreamWriter writer(&serialized);
            toCborStream(writer, object);

            QByteArray compressed;
            gzip::Compress(
                reinterpret_cast<uint8_t*>(serialized.data()), serialized.size(), compressed);
            b.write(compressed);
        }
        else
        {
            QCborStreamWriter writer(&b);
            toCborStream(writer, object);
        }
#else
        Log(QtFatalMsg, FileLogContext, "OAPI_CBOR_SUPPORT is disabed");
        Q_ASSERT(false);
#endif
    }
    break;
    default:
        Log(QtFatalMsg, FileLogContext, "Unknown serialization format");
        Q_ASSERT(false);
        return QByteArray();
    }

    return a;
}

template <typename T>
std::optional<Error> Deserialize(T& out_value, const QByteArray& frame)
{
    auto header = GetHeader(frame);

    QByteArray body =
        QByteArray::fromRawData(frame.data() + sizeof(Header), frame.size() - sizeof(Header));

    if (header->gzip_compressed)
    {
        QByteArray decompressed;
        if (!gzip::Decompress(reinterpret_cast<uint8_t*>(body.data()), body.size(), decompressed))
        {
            const Error e =
                QString("Decompression failed %1").arg(QString::fromLatin1(body.toHex()));
            Log(QtFatalMsg, FileLogContext, e);
            Q_ASSERT(false);
            return e;
        }
        swap(body, decompressed);
    }

    std::optional<Error> error;

    switch (header->serialization_format)
    {
    case oapi::SerializationFormat::E_JSON:
        error = fromJson(out_value, body);
        break;
    case oapi::SerializationFormat::E_CBOR:
#ifdef OAPI_CBOR_SUPPORT
        error = fromCbor(out_value, body);
        break;
#else
        error = Error("OAPI_CBOR_SUPPORT is disabed");
        break;
#endif
    default:
        error = Error("Unknown serialization format");
        break;
    }

    if (error)
    {
        *error += " in: " + (header->serialization_format == oapi::SerializationFormat::E_CBOR
                                 ? body.toHex()
                                 : QString::fromUtf8(body));
        Log(QtFatalMsg, FileLogContext, *error);
        Q_ASSERT(false);
        return error;
    }

    return std::nullopt;
}

}  // namespace streaming_api

#endif
