#include "stream.h"

#include <oapi_models/CommonFrameType.h>
#include <utils/logger.h>
#include <utils/math.h>
#include <utils/thread_name.h>
#include <utils/type_helpers.h>
#include <utils/utils.h>

#include "frame.h"

static const int kMaxNetworkDelayMS = 5000;

namespace streaming_api {

QByteArray StreamConfigFrame(const oapi::AllStreamRequests& request)
{
    return Serialize(request,
                     Header(oapi::StreamType::E_STREAM_CONFIG,
                            utils::to_underlying(oapi::CommonFrameType::E_DATA)));
}

Stream::Stream(QObject* parent) : QObject(parent)
{
    qRegisterMetaType<QAbstractSocket::SocketState>("QAbstractSocket::SocketState");

    QObject::connect(
        &m_thread, &QThread::started, this, &Stream::threadStarted, Qt::DirectConnection);
    QObject::connect(
        &m_thread, &QThread::finished, this, &Stream::threadStopped, Qt::DirectConnection);

    setPlayback(QSharedPointer<Playback>::create());
}

Stream::~Stream()
{
    m_thread.quit();
    m_thread.wait();
}

void Stream::open(const utils::NetworkAddress& address)
{
    close();

    m_config = Configuration{address};

    m_socket = new QWebSocket();
    QObject::connect(&m_thread, &QThread::finished, m_socket, &QWebSocket::deleteLater);

    QObject::connect(m_socket, &QWebSocket::connected, this, [this] { sendFullConfiguration(); });

    QObject::connect(m_socket, &QWebSocket::connected, this, [this] { m_sync_times.reset(); });
    QObject::connect(m_socket, &QWebSocket::connected, this, &Stream::connected);
    QObject::connect(m_socket, &QWebSocket::disconnected, this, &Stream::disconnected);

    QObject::connect(
        m_socket,
        QOverload<QAbstractSocket::SocketError>::of(&QWebSocket::error),
        this,
        [this](QAbstractSocket::SocketError error)
        {
            Log(QtWarningMsg,
                FileLogContext,
                QString("Stream error: %1 (%2)")
                    .arg(m_socket->errorString(), QVariant::fromValue(error).toString()));
        },
        Qt::DirectConnection);

    QObject::connect(
        m_socket,
        &QWebSocket::stateChanged,
        this,
        [](auto state)
        {
            Log(QtInfoMsg,
                FileLogContext,
                "Stream state changed: " + QVariant::fromValue(state).toString());
        },
        Qt::DirectConnection);

    QObject::connect(
        m_socket, &QWebSocket::stateChanged, this, [this](auto state) { m_state = state; });
    QObject::connect(m_socket, &QWebSocket::stateChanged, this, &Stream::stateChanged);

    QObject::connect(m_socket, &QWebSocket::binaryMessageReceived, this, &Stream::processMessage);

    m_socket->moveToThread(&m_thread);
    m_thread.start();
}

void Stream::close()
{
    m_thread.quit();
    m_thread.wait();
    m_config.reset();
}

void Stream::initWrite(const oapi::WriteConfig& config)
{
    m_read_config.reset();
    m_read_streams.reset();
    m_write_config = config;

    sendFullConfiguration();
}

void Stream::setArchivedStreams(const OAISet<oapi::StreamType>& streams)
{
    if (m_read_config)
    {
        Q_ASSERT(false);
        Log(QtWarningMsg, FileLogContext, "Invalid operation on reading stream");
        return;
    }

    m_archived_streams = streams;

    oapi::AllStreamRequests request;
    request.archived_streams = streams;
    sendMessage(StreamConfigFrame(request));
}

void Stream::initRead(const oapi::ReadConfig& config)
{
    m_write_config.reset();
    m_archived_streams.reset();

    m_read_config = config;
    sendFullConfiguration();
}

void Stream::readStreams(const OAISet<oapi::ReadStreamConfig>& streams)
{
    if (m_write_config)
    {
        Q_ASSERT(false);
        Log(QtWarningMsg, FileLogContext, "Invalid operation on writing stream");
        return;
    }

    m_read_streams = streams;

    oapi::AllStreamRequests request;
    request.read_streams = streams;
    sendMessage(StreamConfigFrame(request));
}

void Stream::appendReadStreams(const OAISet<oapi::ReadStreamConfig>& streams)
{
    auto combined = m_read_streams.value_or(OAISet<oapi::ReadStreamConfig>());
    combined.insert(streams.begin(), streams.end());
    readStreams(combined);
}
void Stream::removeReadStreams(const OAISet<oapi::StreamType>& streams)
{
    auto list = m_read_streams.value_or(OAISet<oapi::ReadStreamConfig>());
    utils::erase_if(list, [streams](auto& v) { return streams.find(v.type) != streams.end(); });
    readStreams(list);
}

void Stream::sendSpeed(const double speed)
{
    if (m_write_config)
    {
        Q_ASSERT(false);
        Log(QtWarningMsg, FileLogContext, "Invalid operation on writing stream");
        return;
    }

    oapi::AllStreamRequests request;
    request.speed = speed;
    sendMessage(StreamConfigFrame(request));
}
void Stream::sendSeekTime(const QDateTime& time)
{
    if (m_write_config)
    {
        Q_ASSERT(false);
        Log(QtWarningMsg, FileLogContext, "Invalid operation on writing stream");
        return;
    }

    oapi::AllStreamRequests request;
    request.seek_to = time;
    sendMessage(StreamConfigFrame(request));
}

void Stream::sendMessage(QByteArray&& message)
{
    if (!m_socket)
    {
        return;
    }

    QMetaObject::invokeMethod(m_socket,
                              [this, message = std::move(message)]
                              { m_socket->sendBinaryMessage(std::move(message)); });
}

void Stream::sendMessageSync(QByteArray&& message)
{
    if (!m_socket)
    {
        return;
    }

    QMetaObject::invokeMethod(
        m_socket,
        [this, message = std::move(message)]
        {
            m_socket->sendBinaryMessage(std::move(message));
            m_socket->flush();
        },
        Qt::BlockingQueuedConnection);
}

QSharedPointer<Playback> Stream::playback()
{
    return m_playback;
}

void Stream::setPlayback(const QSharedPointer<Playback>& playback)
{
    if (!playback)
    {
        return;
    }

    if (m_playback)
    {
        QObject::disconnect(m_playback.get(), nullptr, this, nullptr);
    }

    m_playback = playback;
    QObject::connect(m_playback.get(), &Playback::onSeek, this, &Stream::sendSeekTime);
    QObject::connect(m_playback.get(), &Playback::onSpeedChanged, this, &Stream::sendSpeed);

    if (m_read_config)
    {
        sendSeekTime(m_playback->time());
        sendSpeed(m_playback->speed());
    }
}

bool Stream::checkSynchronization(const Header* header)
{
    const auto frame_time = QDateTime::fromMSecsSinceEpoch(header->time, Qt::UTC);

    if (header->frame_type == utils::to_underlying(oapi::CommonFrameType::E_FORMAT))
    {
        // format frames usually created once and long time ago, do not use them for synchronization
        return true;
    }

    if (!m_sync_times)
    {
        m_sync_times = TimeSyncroStruct{frame_time, QDateTime::currentDateTimeUtc(), false};
    }
    else
    {
        // ignore incoming frames, socket will be reinitialized
        if (m_sync_times->aborted)
        {
            return false;
        }
    }

    auto local_dt = m_sync_times->local_time.msecsTo(QDateTime::currentDateTimeUtc());
    auto server_dt = m_sync_times->frame_time.msecsTo(frame_time);
    auto network_delay = local_dt - server_dt;

    if (network_delay < 0)  // newtwork delay reduced and become less 0, resync
    {
        m_sync_times = TimeSyncroStruct{frame_time, QDateTime::currentDateTimeUtc(), false};
        network_delay = 0;
    }

    if (network_delay > kMaxNetworkDelayMS)
    {
        Log(QtWarningMsg,
            FileLogContext,
            QString("Stream time outdated, disconnect: delay %1, local_dt %2, server_dt %3, stream "
                    "type %4, frame type %5")
                .arg(QString::number(network_delay),
                     QString::number(local_dt),
                     QString::number(server_dt),
                     oapi::toJson(header->stream_type),
                     oapi::toJson(header->frame_type)));

        m_sync_times->aborted = true;
        QMetaObject::invokeMethod(m_socket, [this] { m_socket->abort(); });
    }

    return true;
}

void Stream::processMessage(const QByteArray& message)
{
    const auto header = GetHeader(message);
    if (!header)
    {
        return;
    }

    if (checkSynchronization(header))
    {
        emit onMessage(message);
    }
}

void Stream::threadStarted()
{
    utils::SetThreadName("Stream");

    Log(QtInfoMsg, FileLogContext, QString("Starting Stream %1").arg(m_config->address.toString()));

    auto connect = [this]
    {
        m_socket->open(QUrl("ws://" + m_config->address.toString()));
    };

    QObject::connect(
        m_socket,
        &QWebSocket::disconnected,
        this,
        [this, connect] { QTimer::singleShot(1000, m_socket, connect); },
        Qt::DirectConnection);
    connect();
}

void Stream::threadStopped()
{
    Log(QtInfoMsg, FileLogContext, QString("Stopping Stream %1").arg(m_config->address.toString()));

    // was destructed by thread
    m_socket = nullptr;
}

void Stream::sendFullConfiguration()
{
    if (m_read_config)
    {
        oapi::AllStreamRequests config;
        config.read_config = m_read_config;
        config.read_streams = m_read_streams;
        config.seek_to = m_playback->time();
        config.speed = m_playback->speed();
        sendMessage(StreamConfigFrame(config));
    }

    if (m_write_config)
    {
        oapi::AllStreamRequests config;
        config.write_config = m_write_config;
        config.archived_streams = m_archived_streams;
        sendMessage(StreamConfigFrame(config));
    }
}

QAbstractSocket::SocketState Stream::state()
{
    return m_state;
}

}  // namespace streaming_api
