#include "converters.h"

namespace utils {

double BytesPerSecToMBits(const double bytes_sec)
{
    return bytes_sec * 8.0 / kMegaByte;
}

QString BytesPerSecToString(const double bytes_sec)
{
    const double bits_sec = bytes_sec * 8.0;

    QString format;
    double speed = bits_sec;

    if (bits_sec < kKiloByte)
    {
        format = QObject::tr("Bit/s");
    }
    else if (bits_sec < kMegaByte)
    {
        speed = bits_sec / kKiloByte;
        format = QObject::tr("Kbit/s");
    }
    else
    {
        speed = bits_sec / kMegaByte;
        format = QObject::tr("Mbit/s");
    }

    return QString("%1 %2").arg(QString::number(speed, 'f', 2), format);
}

QString BytesToString(const size_t bytes)
{
    QString format;
    double size = static_cast<double>(bytes);

    if (bytes < kKiloByte)
    {
        format = "Bytes";
    }
    else if (bytes < kMegaByte)
    {
        size = size / kKiloByte;
        format = "KBytes";
    }
    else if (bytes < kGigaByte)
    {
        size = size / kMegaByte;
        format = "MBytes";
    }
    else
    {
        size = size / kGigaByte;
        format = "GBytes";
    }

    return QString("%1 %2").arg(QString::number(size, 'f', 2), format);
}

QString MSecsToString(qint64 msecs)
{
    if (msecs < 0)
    {
        msecs *= -1;
    }

    static const qint64 kMsecsPerDay = 1000 * 60 * 60 * 24;

    int day_msecs = msecs % kMsecsPerDay;
    auto days = msecs / kMsecsPerDay;

    QStringList l;
    if (days)
    {
        l.append(QString::number(days) + "d");
    }

    l.append(QTime::fromMSecsSinceStartOfDay(day_msecs).toString("hh:mm:ss.zzz"));
    return l.join(",");
}

QDateTime FromUtcTimeString(const QString& s)
{
    QDateTime t = QDateTime::fromString(s, Qt::ISODateWithMs);
    t.setTimeSpec(Qt::UTC);
    return t;
}

QString BoolToStr(const bool val)
{
    return val ? kTrueStr : kFalseStr;
}

std::set<QString> ToSet(const QStringList& list)
{
    std::set<QString> set;
    for (auto& v : list)
    {
        set.insert(v);
    }

    return set;
}

std::set<size_t> ToIdSet(const QString& str_set)
{
    std::set<size_t> set;
    for (auto& s : str_set.split(", "))
    {
        bool ok = false;
        auto v = s.toUInt(&ok);
        if (ok)
        {
            set.insert(v);
        }
    }

    return set;
}

std::optional<qint32> ToInt32(const QVariant& value)
{
    if (value.isNull() || !value.isValid())
    {
        return {};
    }

    bool ok = false;
    int res = value.toInt(&ok);
    return ok ? res : std::optional<qint32>();
}

std::optional<qint64> ToInt64(const QVariant& value)
{
    if (value.isNull() || !value.isValid())
    {
        return {};
    }

    bool ok = false;
    int res = value.toLongLong(&ok);
    return ok ? res : std::optional<qint64>();
}

std::optional<float> ToFloat(const QVariant& value)
{
    if (value.isNull() || !value.isValid())
    {
        return {};
    }

    bool ok = false;
    float res = value.toFloat(&ok);
    return ok ? res : std::optional<float>();
}

std::optional<double> ToDouble(const QVariant& value)
{
    if (value.isNull() || !value.isValid())
    {
        return {};
    }

    bool ok = false;
    float res = value.toDouble(&ok);
    return ok ? res : std::optional<double>();
}

std::optional<bool> ToBool(const QVariant& value)
{
    if (value.isNull() || !value.isValid())
    {
        return {};
    }

    auto str = value.toString().toLower();
    bool is_true = str == kTrueStr.toLower();
    bool is_false = str == kFalseStr.toLower();

    if (!is_true && !is_false)
    {
        return {};
    }

    return is_true;
}

std::optional<QDateTime> ToQDateTime(const QVariant& value)
{
    auto dt = value.toDateTime();
    if (dt.isNull() || !dt.isValid())
    {
        return std::nullopt;
    }
    return dt;
}

bool CanConvert(const QVariant& value, const QVariant::Type type)
{
    switch (static_cast<int>(type))
    {
    case QVariant::Int:
        return ToInt32(value).has_value();
    case QVariant::LongLong:
        return ToInt64(value).has_value();
    case QVariant::Double:
        return ToDouble(value).has_value();
    case QMetaType::Float:
        return ToFloat(value).has_value();
    case QVariant::Bool:
        return ToBool(value).has_value();
    case QVariant::String:
        return !value.isNull() && value.isValid();
    case QVariant::DateTime:
        return ToQDateTime(value).has_value();
    default:
        return false;
    }
}

void OptimizeQVariantType(QVariant& value)
{
    if (CanConvert(value, QVariant::Int))
    {
        value = value.toInt();
        return;
    }

    if (CanConvert(value, QVariant::Double))
    {
        value = value.toDouble();
        return;
    }

    if (CanConvert(value, QVariant::Bool))
    {
        value = value.toBool();
        return;
    }
}

QString IntToHexString(const int32_t value)
{
    return QString("%1").arg(value, 8, 16, QLatin1Char('0')).toUpper();
}

}  // namespace utils
