#include "gzip_compressor.h"

namespace gzip {

const int GZIP_WINDOWS_BIT = 15 + 16;
const int GZIP_CHUNK_SIZE = 32 * 1024;

template <typename T>
bool Compress(const uint8_t* input, const size_t input_size, T& output, const int level)
{
    // Is there something to do?
    if (!input_size)
    {
        return (false);
    }

    // Declare vars
    int flush = 0;

    // Prepare deflater status
    z_stream strm;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    strm.avail_in = 0;
    strm.next_in = Z_NULL;

    // Initialize deflater
    int ret = deflateInit2(
        &strm, qMax(-1, qMin(9, level)), Z_DEFLATED, GZIP_WINDOWS_BIT, 8, Z_DEFAULT_STRATEGY);

    if (ret != Z_OK)
    {
        return (false);
    }

    // Extract pointer to input data
    uint8_t* input_data = const_cast<uint8_t*>(input);
    int input_data_left = input_size;

    // Compress data until available
    do
    {
        // Determine current chunk size
        int chunk_size = qMin(GZIP_CHUNK_SIZE, input_data_left);

        // Set deflater references
        strm.next_in = reinterpret_cast<unsigned char*>(input_data);
        strm.avail_in = chunk_size;

        // Update interval variables
        input_data += chunk_size;
        input_data_left -= chunk_size;

        // Determine if it is the last chunk
        flush = (input_data_left <= 0 ? Z_FINISH : Z_NO_FLUSH);

        // Deflate chunk and cumulate output
        do
        {
            // Set inflater references
            auto pos = output.size();
            output.resize(pos + GZIP_CHUNK_SIZE);

            strm.next_out = reinterpret_cast<unsigned char*>(output.data() + pos);
            strm.avail_out = GZIP_CHUNK_SIZE;

            // Try to deflate chunk
            ret = deflate(&strm, flush);

            // Check errors
            if (ret == Z_STREAM_ERROR)
            {
                // Clean-up
                deflateEnd(&strm);

                // Return
                return (false);
            }

            if (strm.avail_out)
            {
                output.resize(output.size() - strm.avail_out);
            }
        }
        while (strm.avail_out == 0);
    }
    while (flush != Z_FINISH);

    // Clean-up
    (void)deflateEnd(&strm);

    // Return
    return (ret == Z_STREAM_END);
}

template <typename T>
bool Decompress(const uint8_t* input, const size_t input_size, T& output)
{
    if (!input_size)
    {
        return false;
    }

    // Prepare inflater status
    z_stream strm;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    strm.avail_in = 0;
    strm.next_in = Z_NULL;

    // Initialize inflater
    int ret = inflateInit2(&strm, GZIP_WINDOWS_BIT);

    if (ret != Z_OK)
    {
        return (false);
    }

    // Extract pointer to input data
    uint8_t* input_data = const_cast<uint8_t*>(input);
    int input_data_left = input_size;

    // Decompress data until available
    do
    {
        // Determine current chunk size
        int chunk_size = qMin(GZIP_CHUNK_SIZE, input_data_left);

        // Check for termination
        if (chunk_size <= 0)
        {
            break;
        }

        // Set inflater references
        strm.next_in = reinterpret_cast<unsigned char*>(input_data);
        strm.avail_in = chunk_size;

        // Update interval variables
        input_data += chunk_size;
        input_data_left -= chunk_size;

        // Inflate chunk and cumulate output
        do
        {
            // Set inflater references
            auto pos = output.size();
            output.resize(pos + GZIP_CHUNK_SIZE);

            strm.next_out = reinterpret_cast<unsigned char*>(output.data() + pos);
            strm.avail_out = GZIP_CHUNK_SIZE;

            // Try to inflate chunk
            ret = inflate(&strm, Z_NO_FLUSH);

            if (ret == Z_NEED_DICT)
            {
                ret = Z_DATA_ERROR;
            }

            switch (ret)
            {
            case Z_DATA_ERROR:
            case Z_MEM_ERROR:
            case Z_STREAM_ERROR:
                // Clean-up
                inflateEnd(&strm);
                return (false);
            }

            if (strm.avail_out)
            {
                output.resize(output.size() - strm.avail_out);
            }
        }
        while (strm.avail_out == 0);
    }
    while (ret != Z_STREAM_END);

    // Clean-up
    inflateEnd(&strm);

    // Return
    return (ret == Z_STREAM_END || ret == Z_OK);
}

// manually declare available template types
auto cv = Compress<std::vector<uint8_t>>;
auto cb = Compress<QByteArray>;

auto dv = Decompress<std::vector<uint8_t>>;
auto db = Decompress<QByteArray>;

}  // namespace gzip
