#include "logger.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDir>
#include <QFile>
#include <QStandardPaths>
#include <cstdarg>
#include <cstdio>
#include <iostream>
#include <mutex>
#include <shared_mutex>
#include <sstream>

#include "os.h"
#include "utils.h"

QLatin1String Logger::kAllCategory("all");

bool IsLog(QtMsgType type, const LogContextEx& context)
{
    auto logger = LoggerInstance::Instance();
    return logger && logger->isEnabledType(type) && logger->isEnabledCategory(context.m_category);
}

void Log(QtMsgType type, const QMessageLogContext& context, const QString& msg)
{
    Log(type, LogContextEx(context), msg);
}

void Log(QtMsgType type, const LogContextEx& context, const QString& msg)
{
    auto logger = LoggerInstance::Instance();
    if (!logger)
    {
        switch (type)
        {
        case QtDebugMsg:
            QMessageLogger(context.m_file, context.m_line, context.m_function).debug() << msg;
            return;
        case QtWarningMsg:
            QMessageLogger(context.m_file, context.m_line, context.m_function).warning() << msg;
            return;
        case QtCriticalMsg:
        case QtFatalMsg:
            QMessageLogger(context.m_file, context.m_line, context.m_function).critical() << msg;
            return;
        case QtInfoMsg:
            QMessageLogger(context.m_file, context.m_line, context.m_function).info() << msg;
            return;
        default:
            return;
        }
    }

    QMetaObject::invokeMethod(logger.get(),
                              [logger, type, msg, context] { logger->log(type, context, msg); });
}

Logger::Logger()
{
    qRegisterMetaType<QtMsgType>("QtMsgType");
    qRegisterMetaType<LogContextEx>("LogContextEx");

    Q_ASSERT(QCoreApplication::instance());

    m_previous_handler = qInstallMessageHandler(Log);
    setDiskConfiguration(QStandardPaths::writableLocation(QStandardPaths::AppLocalDataLocation) +
                         "/Logs");

    m_known_categories.insert(kAllCategory);
}

void Logger::setDiskConfiguration(const QString& directory, qint64 max_file_size)
{
    m_directory = directory;
    m_max_file_size = max_file_size;
    m_file.close();
}

void Logger::setDepth(const size_t depth_hours)
{
    m_remove_timer.stop();
    m_remove_timer.disconnect();
    m_depth_hours = depth_hours;

    if (depth_hours)
    {
        QObject::connect(&m_remove_timer, &QTimer::timeout, this, &Logger::removeDeprecatedLogs);
        m_remove_timer.start(1000 * 60 * 60);
    }
    removeDeprecatedLogs();
}

void Logger::removeDeprecatedLogs()
{
    QDir dir(m_directory);
    if (!dir.exists())
    {
        return;
    }

    foreach (auto file, dir.entryInfoList(QStringList("*.txt"), QDir::Files))
    {
        if (file.birthTime().addSecs(m_depth_hours * 60 * 60) < QDateTime::currentDateTime())
        {
            if (m_file.fileName() != file.absoluteFilePath())
            {
                QFile::remove(file.absoluteFilePath());
            }
        }
    }
}

void Logger::setEnabledCategories(const std::set<QString>& categories)
{
    m_enabled_categories = categories;
}

void Logger::enableCategory(const QString& category)
{
    m_enabled_categories.insert(category);
}

void Logger::disableCategory(const QString& category)
{
    m_enabled_categories.erase(category);
}

bool Logger::isEnabledCategory(const QString& category)
{
    bool known = (m_known_categories.find(category) != m_known_categories.end());

    if (!known)
    {
        m_known_categories.insert(category);
        emit onNewCategory(category);
    }

    return (m_enabled_categories.find(category) != m_enabled_categories.end()) ||
           (m_enabled_categories.find(kAllCategory) != m_enabled_categories.end());
}

std::set<QString> Logger::getKnownCategories()
{
    return m_known_categories;
}

std::set<QString> Logger::getEnabledCategories()
{
    return m_enabled_categories;
}

void Logger::setEnabledTypes(const std::set<QtMsgType>& types)
{
    m_enabled_types = types;
}

void Logger::enableType(const QtMsgType type)
{
    m_enabled_types.insert(type);
}

void Logger::disableType(const QtMsgType type)
{
    m_enabled_types.erase(type);
}

bool Logger::isEnabledType(const QtMsgType type)
{
    return m_enabled_types.find(type) != m_enabled_types.end();
}

std::set<QtMsgType> Logger::getEnabledTypes()
{
    return m_enabled_types;
}

QString CreateMessage(QtMsgType type, const LogContextEx& context, const QString& msg)
{
    const char* message_type = "";

    switch (type)
    {
    case QtInfoMsg:
        message_type = "Info";
        break;
    case QtDebugMsg:
        message_type = "Debug";
        break;
    case QtWarningMsg:
        message_type = "Warning";
        break;
    case QtCriticalMsg:
        message_type = "CRITICAL";
        break;
    case QtFatalMsg:
        message_type = "FATAL";
        break;
    }

    return QString("[%1] %2 - %3: %4")
        .arg(QDateTime::currentDateTimeUtc().toString(Qt::ISODateWithMs),
             context.m_category,
             message_type,
             msg);
}

void Logger::log(QtMsgType type, const LogContextEx& context, const QString& msg)
{
    emit onUnfilteredLog(type, context, msg);
    if (!isEnabledType(type) || !isEnabledCategory(context.m_category))
    {
        return;
    }

    auto message = CreateMessage(type, context, msg);

    if (m_file.size() > m_max_file_size)
    {
        m_file.close();
    }

    if (!m_file.isOpen())
    {
        if (!QDir().exists(m_directory))
        {
            utils::CreatePathWithUserPermissions(m_directory);
        }

        auto file_path =
            QString("%1/log-%2.txt")
                .arg(m_directory, QDateTime::currentDateTimeUtc().toString("yyyy-MM-dd-HH-mm-ss"));

        m_file.setFileName(file_path);
        m_file.open(QFile::WriteOnly);
    }

    if (m_file.isOpen())
    {
        m_file.write(message.toUtf8());
        m_file.write("\n");
    }

    if (type == QtInfoMsg)
    {
        std::clog << message.toStdString() << "\n";
    }
    else
    {
        std::cerr << message.toStdString() << "\n";
    }

    emit onLog(type, context, message);
}

Logger::~Logger()
{
    qInstallMessageHandler(m_previous_handler);
}
