#pragma once

#ifndef LOGGER_H
#define LOGGER_H

#include <QDebug>
#include <QFile>
#include <QFileInfo>
#include <QLoggingCategory>
#include <QTimer>
#include <set>
#include <shared_mutex>

#include "singleton.h"

#ifdef QT_MESSAGELOGCONTEXT

namespace logger_utils {

// recursive compile-time functions
int constexpr StrLength(const char* str)
{
    return *str ? 1 + StrLength(str + 1) : 0;
}

constexpr int GetFileNameIndex(const char path[],
                               const int32_t index = 0,
                               const int32_t slash_index = -1)
{
    return path[index] ? ((path[index] == '/' || path[index] == '\\')  // (see below)
                              ? GetFileNameIndex(path, index + 1, index)
                              : GetFileNameIndex(path, index + 1, slash_index))
                       : (slash_index + 1);
}

constexpr bool IsExtensionIndex(const char path[], const int32_t index = 0)
{
    return path[index] == '.' && (path[index + 1] == 'h' || path[index + 1] == 'c');
}

constexpr int GetExtensionIndex(const char path[], const int32_t index = 0)
{
    return IsExtensionIndex(path, index) || !path[index] ? index
                                                         : GetExtensionIndex(path, index + 1);
}

// generate id from file & line, so every SubstringOffsets instance is unique
int constexpr StructId(const char* file, int line)
{
    return StrLength(file) + line;
}

template <int struct_id, int start, int end>
struct SubstringOffsets
{
    static constexpr int kLength = end - start;

    // fast getting QLatin1String substring from pre-calculated offsets
    static QLatin1String str(const char* base_str)
    {
        return QLatin1String(base_str + start, kLength);
    }
};

}  // namespace logger_utils

#define BASE_FILENAME                                                                              \
    logger_utils::SubstringOffsets<logger_utils::StructId(QT_MESSAGELOG_FILE, QT_MESSAGELOG_LINE), \
                                   logger_utils::GetFileNameIndex(QT_MESSAGELOG_FILE),             \
                                   logger_utils::GetExtensionIndex(                                \
                                       QT_MESSAGELOG_FILE)>::str(__FILE__)

#define FileLogContext \
    LogContextEx(QT_MESSAGELOG_FILE, QT_MESSAGELOG_LINE, QT_MESSAGELOG_FUNC, BASE_FILENAME)
#define LogContext(context) \
    LogContextEx(QT_MESSAGELOG_FILE, QT_MESSAGELOG_LINE, QT_MESSAGELOG_FUNC, context)
#else
#define FileLogContext LogContextEx("unknown", 0, "unknown", "default")
#define LogContext(context) LogContextEx("unknown", 0, "unknown", context)
#endif

class LogContextEx
{
public:
    LogContextEx() = default;
    constexpr LogContextEx(const QMessageLogContext& context)
            : m_line(context.line)
            , m_file(context.file)
            , m_function(context.function)
            , m_category(context.category)
    {
    }

    constexpr LogContextEx(const char* file,
                           int line,
                           const char* function,
                           const QLatin1String& category) noexcept
            : m_line(line), m_file(file), m_function(function), m_category(category)
    {
    }

    constexpr LogContextEx(const char* file,
                           int line,
                           const char* function,
                           const char* category) noexcept
            : m_line(line), m_file(file), m_function(function), m_category(category)
    {
    }

    int m_line = 0;
    const char* m_file = nullptr;
    const char* m_function = nullptr;
    const QLatin1String m_category;
};

bool IsLog(QtMsgType type, const LogContextEx& context);
void Log(QtMsgType type, const LogContextEx& context, const QString& msg);
void Log(QtMsgType type, const QMessageLogContext& context, const QString& msg);

class Logger : public QObject
{
    Q_OBJECT
public:
    static QLatin1String kAllCategory;

    friend void Log(QtMsgType type, const QMessageLogContext& context, const QString& msg);

    Logger();
    ~Logger() override;

    void setDiskConfiguration(const QString& directory, qint64 max_file_size = kMaxFileSize);

    void setDepth(const size_t depth_hours);

    void setEnabledCategories(const std::set<QString>& categories);
    void enableCategory(const QString& category);
    void disableCategory(const QString& category);
    bool isEnabledCategory(const QString& category);
    std::set<QString> getKnownCategories();
    std::set<QString> getEnabledCategories();

    void setEnabledTypes(const std::set<QtMsgType>& types);
    void enableType(const QtMsgType type);
    void disableType(const QtMsgType type);
    bool isEnabledType(const QtMsgType type);
    std::set<QtMsgType> getEnabledTypes();

    void log(QtMsgType type, const LogContextEx& context, const QString& msg);

signals:
    void onNewCategory(const QString& category);
    void onLog(QtMsgType type, const LogContextEx& context, const QString& msg);
    void onUnfilteredLog(QtMsgType type, const LogContextEx& context, const QString& msg);

private slots:
    void removeDeprecatedLogs();

private:
    static const qint64 kMaxFileSize = 1024 * 1024;

    std::set<QString> m_known_categories;
    std::set<QString> m_enabled_categories = {kAllCategory};
    std::set<QtMsgType> m_enabled_types = {
        QtInfoMsg, QtDebugMsg, QtWarningMsg, QtCriticalMsg, QtFatalMsg};

    QString m_directory;
    qint64 m_max_file_size;

    QTimer m_remove_timer;
    size_t m_depth_hours;

    QFile m_file;

    QtMessageHandler m_previous_handler;
};

using LoggerInstance = utils::Singleton<Logger>;

#endif
