#ifndef UTILS_MATH_H
#define UTILS_MATH_H

#define _USE_MATH_DEFINES
#ifndef NOMINMAX
#define NOMINMAX
#endif

#undef min
#undef max

#include <math.h>

#include <algorithm>
#include <cmath>
#include <cstddef>

#include "type_helpers.h"

namespace utils {

constexpr float kPIf = static_cast<float>(M_PI);
constexpr float kMegaScale = 1000000.0f;
constexpr float kGigaScale = 1000000000.0f;

constexpr float kMicroScale = 1.f / kMegaScale;
constexpr float kNanoScale = 1.f / kGigaScale;

constexpr float kDegreeBand = 360.f;
constexpr float kLightSpeed = 3e8f;

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
inline T RadToDeg(T angle)
{
    return angle * static_cast<T>(180) / static_cast<T>(M_PI);
}

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
inline T DegToRad(T angle)
{
    return angle * static_cast<T>(M_PI) / static_cast<T>(180);
}

/**
 * Make sure that that the distance between two angles is the smallest one.
 * For example for the input (359, 1, 180) the output will be (359, 361).
 * pi should be in the same units as the angles.
 * Both angles should be from 0 to pi * 2.
 */
template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
void FixRelativeAnglesPair(const T& lhs, T& rhs, const T pi)
{
    T two_pi = pi * 2;

    if (rhs > lhs + pi)
    {
        rhs -= two_pi;
    }
    else if (rhs < lhs - pi)
    {
        rhs += two_pi;
    }
}

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
T NormalizeAngle(T angle, const T pi)
{
    T two_pi = pi * 2;

    if (angle < 0.0)
    {
        angle += std::ceil(-angle / two_pi) * two_pi;
        return angle;
    }

    if (angle > two_pi)
    {
        angle -= std::floor(angle / two_pi) * two_pi;
    }
    return angle;
}

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
T NormalizeDegrees(T degrees)
{
    return NormalizeAngle(degrees, T(180));
}

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
T NormalizeRadians(T radians)
{
    return NormalizeAngle(radians, M_PI);
}

template <typename T>
inline T NormalizeMinMax(const T& value, const T& min_value, const T& max_value)
{
    if (max_value == min_value)
    {
        return 0;
    }

    T in_range = std::min<T>(max_value, std::max<T>(min_value, value));
    return std::min<T>(
        static_cast<T>(1),
        std::max<T>(static_cast<T>(0), (in_range - min_value) / (max_value - min_value)));
}

/**
 * If the frames number is incremental but can overflow we can use this method to find
 * the frame number difference.
 */
template <typename T, utils::enable_if<std::is_unsigned<T>::value> = 0>
T SubOverflow(T lhs, T rhs, T maxValue = std::numeric_limits<T>::max())
{
    return lhs >= rhs ? lhs - rhs : (maxValue - (rhs - lhs) + 1);
}

template <typename T>
inline T AbsDiff(T a, T b)
{
    return a > b ? a - b : b - a;
}

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
inline bool FpEqual(const T lhs, const T rhs, T eps = std::numeric_limits<T>::epsilon())
{
    return std::abs(lhs - rhs) <= eps;
}

template <typename T,
          typename SignedT = typename std::make_signed<T>::type,
          utils::enable_if<std::is_integral<T>::value> = 0>
SignedT CircularRange(T lhs, T rhs, T period)
{
    const auto sub = static_cast<SignedT>(rhs) - static_cast<SignedT>(lhs);
    if (std::abs(sub) <= (static_cast<SignedT>(period) / static_cast<SignedT>(2)))
    {
        return sub;
    }

    return sub < 0 ? (static_cast<SignedT>(period) - std::abs(sub))
                   : -(static_cast<SignedT>(period) - std::abs(sub));
}

template <typename T, utils::enable_if<std::is_floating_point<T>::value> = 0>
T CircularRangeFP(const T& lhs, const T& rhs, const T& period)
{
    const auto sub = rhs - lhs;
    if (std::abs(sub) <= period / static_cast<T>(2))
    {
        return sub;
    }

    return sub < 0 ? (period - std::abs(sub)) : -(period - std::abs(sub));
}

template <typename T>
T Square(T rhs)
{
    return rhs * rhs;
}

template <typename T>
inline T UpperLog2(T value)
{
    return static_cast<T>(std::ceil(std::log2(static_cast<double>(value))));
}

template <typename T>
inline T ToDb(T value)
{
    return static_cast<T>(10) * log10(value);
}

template <typename T>
inline T FromDb(T value)
{
    return pow(static_cast<T>(10), value / static_cast<T>(10));
}

template <typename T>
inline T UpperPow2(T value)
{
    return T(1) << UpperLog2(value);
}

template <typename T>
inline T ClosestPow2(T value)
{
    T upper_pow = UpperLog2(value);

    T upper_val = T(1) << upper_pow;

    if (upper_pow <= T(1))
    {
        return upper_val;
    }

    T lower_pow = upper_pow - T(1);
    T lower_val = T(1) << lower_pow;

    if (value - lower_val < upper_val - value)
    {
        return lower_val;
    }

    return upper_val;
}

}  // namespace utils

#endif
