#include "network.h"

#include "logger.h"

namespace utils {

const QString kIPRangeValidationRange = "(?:[0-1]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])";
const QString kIPValidationRegex = "^" + kIPRangeValidationRange + "\\." + kIPRangeValidationRange +
                                   "\\." + kIPRangeValidationRange + "\\." +
                                   kIPRangeValidationRange + "$";

const QString kMacValidationRegex = "^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$";

QString IpAddress::toString() const
{
    return QString("%1.%2.%3.%4")
        .arg(QString::number(v[0]),
             QString::number(v[1]),
             QString::number(v[2]),
             QString::number(v[3]));
}

bool IpAddress::fromString(const QString& str)
{
    static const int kIpElements = 4;
    QStringList elements = str.split(QString("."));
    if (elements.size() != kIpElements)
    {
        return false;
    }

    IpAddress ip;
    for (int i = 0; i < kIpElements; ++i)
    {
        bool ok = true;
        auto val = elements[i].toInt(&ok);
        if (!ok || val < 0 || val > 255)
        {
            return false;
        }

        ip.v[i] = static_cast<uint8_t>(val);
    }
    *this = ip;
    return true;
}

QString MacAddress::toString() const
{
    return QString("%1:%2:%3:%4:%5:%6")
        .arg(QString("%1").arg(v[0], 2, 16, QLatin1Char('0')).toUpper(),
             QString("%1").arg(v[1], 2, 16, QLatin1Char('0')).toUpper(),
             QString("%1").arg(v[2], 2, 16, QLatin1Char('0')).toUpper(),
             QString("%1").arg(v[3], 2, 16, QLatin1Char('0')).toUpper(),
             QString("%1").arg(v[4], 2, 16, QLatin1Char('0')).toUpper(),
             QString("%1").arg(v[5], 2, 16, QLatin1Char('0')).toUpper());
}

bool MacAddress::fromString(const QString& str)
{
    static const int kMacElements = 6;
    QStringList elements = str.split(QString(":"));
    if (elements.size() != kMacElements)
    {
        return false;
    }

    MacAddress ip;
    for (int i = 0; i < kMacElements; ++i)
    {
        bool ok = true;
        auto val = elements[i].toInt(&ok, 16);
        if (!ok || val < 0 || val > 255)
        {
            return false;
        }

        ip.v[i] = static_cast<uint8_t>(val);
    }
    *this = ip;
    return true;
}

QHostAddress GetLocalHostAddress(const utils::NetworkAddress& destination_address)
{
    QTcpSocket socket;
    socket.connectToHost(destination_address.host, destination_address.port);
    socket.waitForConnected(4000);

    auto state = socket.state();
    if (state != QAbstractSocket::ConnectedState)
    {
        return QHostAddress(QHostAddress::Null);
    }

    return socket.localAddress();
}

void GetLocalHostAddressAsync(const utils::NetworkAddress& destination_address,
                              QObject* owner,
                              std::function<void(const QHostAddress&)> callback)
{
    QTcpSocket* socket = new QTcpSocket(owner);
    QTimer* timer = new QTimer(owner);

    QObject::connect(socket,
                     &QTcpSocket::connected,
                     owner,
                     [=]()
                     {
                         socket->deleteLater();
                         timer->deleteLater();
                         callback(socket->localAddress());
                     });

    timer->setSingleShot(true);

    QObject::connect(timer,
                     &QTimer::timeout,
                     owner,
                     [=]()
                     {
                         Log(QtFatalMsg, FileLogContext, socket->errorString());
                         socket->deleteLater();
                         timer->deleteLater();
                         callback(QHostAddress(QHostAddress::Null));
                     });

    socket->connectToHost(destination_address.host, destination_address.port);
    timer->start(4000);
}

QList<QNetworkProxy> GetInternetProxies()
{
    QNetworkProxyQuery npq(QUrl("www.google.com"));
    return QNetworkProxyFactory::systemProxyForQuery(npq);
}

}  // namespace utils
