#include "os.h"
#include <QDir>

#ifdef _WIN32

#ifdef _MSC_VER
#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "shell32.lib")
#endif

// clang-format off
#include <windows.h>
#include <sddl.h>
#include <stdio.h>
// clang-format on

#include "shlobj.h"
#endif

#ifndef _WIN32
#include <sys/stat.h>
#include <sys/types.h>
#endif

#include <QFile>

namespace utils {

#ifdef _WIN32
// CreateMyDACL.
//    Create a security descriptor that contains the DACL
//    you want.
//    This function uses SDDL to make Deny and Allow ACEs.
//
// Parameter:
//    SECURITY_ATTRIBUTES * pSA
//    Pointer to a SECURITY_ATTRIBUTES structure. It is your
//    responsibility to properly initialize the
//    structure and to free the structure's
//    lpSecurityDescriptor member when you have
//    finished using it. To free the structure's
//    lpSecurityDescriptor member, call the
//    LocalFree function.
//
// Return value:
//    FALSE if the address to the structure is NULL.
//    Otherwise, this function returns the value from the
//    ConvertStringSecurityDescriptorToSecurityDescriptor
//    function.
BOOL CreateDACL(SECURITY_ATTRIBUTES* pSA)
{
    // Define the SDDL for the DACL. This example sets
    // the following access:
    //     Built-in guests are denied all access.
    //     Anonymous logon is denied all access.
    //     Authenticated users are allowed
    //     read/write/execute access.
    //     Administrators are allowed full control.
    // Modify these values as needed to generate the proper
    // DACL for your application.
    const TCHAR* szSD = TEXT("D:")  // Discretionary ACL
        TEXT("(D;OICI;GA;;;BG)")    // Deny access to
                                    // built-in guests
        TEXT("(D;OICI;GA;;;AN)")    // Deny access to
                                    // anonymous logon
        TEXT("(A;OICI;GA;;;AU)")    // Allow
                                    // read/write/execute
                                    // to authenticated
                                    // users
        TEXT("(A;OICI;GA;;;BA)");   // Allow full control
                                    // to administrators

    if (NULL == pSA)
    {
        return FALSE;
    }

    return ConvertStringSecurityDescriptorToSecurityDescriptor(
        szSD, SDDL_REVISION_1, &(pSA->lpSecurityDescriptor), NULL);
}

#endif

bool CreatePathWithUserPermissions(QString path)
{
#ifdef _WIN32
    path.replace("/", "\\\\");
    SECURITY_ATTRIBUTES sa;

    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = FALSE;

    // Call function to set the DACL. The DACL
    // is set in the SECURITY_ATTRIBUTES
    // lpSecurityDescriptor member.
    if (!CreateDACL(&sa))
    {
        return false;
    }

    // Use the updated SECURITY_ATTRIBUTES to specify
    // security attributes for securable objects.
    // This example uses security attributes during
    // creation of a new directory.
    if (0 != SHCreateDirectoryExW(NULL, path.toStdWString().c_str(), &sa))
    {
        return false;
    }

    LocalFree(sa.lpSecurityDescriptor);
    return true;
#else
    SetLinuxFilePermissions();
    return QDir().mkpath(path);
#endif
}

bool CreateFileWithUsersPermissions(QString name, const QByteArray& data)
{
#ifdef _WIN32
    name.replace("/", "\\\\");
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = FALSE;

    if (!CreateDACL(&sa))
    {
        return false;
    }

    auto file = CreateFile(name.toStdWString().c_str(),
                           GENERIC_WRITE,
                           FILE_SHARE_WRITE,
                           &sa,
                           CREATE_ALWAYS,
                           FILE_ATTRIBUTE_NORMAL,
                           NULL);

    if (file == INVALID_HANDLE_VALUE)
    {
        return false;
    }

    DWORD dwBytesWritten = 0;
    bool result = WriteFile(file, data.data(), data.size(), &dwBytesWritten, NULL);

    CloseHandle(file);
    return result;
#else
    SetLinuxFilePermissions();
    QFile file(name);
    return file.open(QIODevice::ReadWrite) && file.write(data) == data.size();
#endif
}

#ifndef _WIN32
void SetLinuxFilePermissions()
{
    umask(0000);
}
#endif

bool SetFileUsersPermissions(QString name)
{
    QFile orig_file(name);
    if (!orig_file.open(QFile::ReadOnly))
    {
        return false;
    }

    auto file_data = orig_file.readAll();
    orig_file.close();
    if (!QFile::remove(name))
    {
        return false;
    }

    return CreateFileWithUsersPermissions(name, file_data);
}

}  // namespace utils
