#ifndef RESULT_H
#define RESULT_H

#include <QNetworkReply>
#include <QString>
#include <functional>
#include <memory>
#include <optional>

class Error : public QString
{
public:
    Error() : QString() {}
    Error(const QString& s) : QString(s) {}
    Error(const QLatin1String& s) : QString(s) {}
    Error(const char* s) : QString(s) {}
};

template <typename T, typename E = Error>
struct ResultPtr
{
    ResultPtr(const std::shared_ptr<T>& _result) : result(_result) {}

    ResultPtr(const QString& _error) : error(_error) {}

    ResultPtr(const ResultPtr& _result) : result(_result.result), error(_result.error) {}

    ResultPtr operator=(const ResultPtr& rhs)
    {
        result = rhs.result;
        error = rhs.error;
        return *this;
    }

    std::shared_ptr<T> result;
    std::optional<E> error;
};

template <typename T, typename E = Error>
struct ResultOptional
{
    ResultOptional(const T& _result) : result(_result) {}

    ResultOptional(const E& _error) : error(_error) {}

    ResultOptional(const ResultOptional& _result) : result(_result.result), error(_result.error) {}

    ResultOptional operator=(const ResultOptional& rhs)
    {
        result = rhs.result;
        error = rhs.error;
        return *this;
    }

    std::optional<T> result;
    std::optional<E> error;
};

template <typename E>
struct ResultVoid
{
    ResultVoid() = default;

    ResultVoid(const E& _error) : error(_error) {}

    ResultVoid(const ResultVoid& result) : error(result.error) {}

    ResultVoid operator=(const ResultVoid& rhs)
    {
        error = rhs.error;
        return *this;
    }

    std::optional<E> error;
};

using Result = ResultVoid<Error>;

struct RequestError
{
    RequestError() = default;

    RequestError(const QString& _description) : description(_description) {}

    RequestError(const QString& _description,
                 const QNetworkReply::NetworkError _network_error,
                 const int _http_status_code)
            : description(QString("%2 (%3) %1")
                              .arg(_description,
                                   QVariant::fromValue(_network_error).toString(),
                                   QString::number(_http_status_code)))
            , network_error(_network_error)
            , http_status_code(_http_status_code)
    {
    }

    QString description;
    QNetworkReply::NetworkError network_error = QNetworkReply::NetworkError::NoError;
    int http_status_code = 0;
};

template <typename T>
using RequestResult = ResultOptional<T, RequestError>;

#endif  // RESULT_H
